'''
Functions for analyzing Molang expressions.
'''
import re
from collections import defaultdict

def find_molang_resources(
        molang: str, resource_prefixes: list[str]) -> dict[str, list[str]]:
    '''
    Finds all resources of specified types in a molang expression. Returns
    a dictionary keyed by the resource type, with values being a list of
    found resources.

    Example
    -------
    >>> find_molang_resources(
            molang=(
                "q.is_baby = 1 ? array.skin[q.variant] : ("
                "q.mark_variant ? texture.default : texture.default2)"
            ),
            resource_prefixes=["array", "texture"])
    {'array': ['array.skin'], 'texture': ['texture.default', 'texture.default2']}
    '''
    molang = molang.lower()
    results: dict[str, list[str]] = {
        prefix: []
        for prefix in resource_prefixes
    }
    for resource_prefix in resource_prefixes:
        # Capturing groups crops the first part of the search for example:
        # geometry.default -> geometry
        item_pattern = re.compile(
            f'{resource_prefix}\\.(\\w+)', flags=re.IGNORECASE)
        for item_name in item_pattern.findall(molang):
            results[resource_prefix].append(item_name)
    return results

def split_item_name(item_name: str) -> tuple[str, str, int]:
    '''
    Splits given name of the item into three parts: napespace, name and data
    value. If the name is missing the namespace, than the default namespace
    value is minecraft. If the data value is missing is set to 0.

    In rare cases when the data value can't be converted to int, the data
    value and the separator (:) are considered to be part of the name.

    Example
    -------
    >>> split_item_name("minecraft:stone")
    ('minecraft', 'stone', 0)
    >>> split_item_name("stone")
    ('minecraft', 'stone', 0)
    >>> split_item_name("minecraft:stone:1")
    ('minecraft', 'stone', 1)
    '''
    item_name = item_name.lower()
    if ':' in item_name:
        namespace, name = item_name.split(':', 1)
    else:
        namespace = 'minecraft'
        name = item_name
    if ':' in name:
        name, data = name.rsplit(':', 1)
        try:
            data = int(data)
        except ValueError:
            name = f'{name}:{data}'
            data = 0
    else:
        data = 0
    return namespace, name, data

def parse_format_version(version: str) -> tuple[int, ...]:
    '''
    Parses string with format version (dot separated numbers).
    '''
    return tuple(int(part) for part in version.split('.'))


from __future__ import annotations

import json
import re
from json import JSONDecodeError
from json import scanner  # type: ignore
from json.decoder import WHITESPACE, WHITESPACE_STR, scanstring  # type: ignore

# JSON Decoder
FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
INLINE_COMMENT = re.compile(r'//[^\n]*\n?', FLAGS)
INLINE_COMMENT_STRING_START = '//'
MULTILINE_COMMENT = re.compile(r"/[*]([^*]|([*][^/]))*[*]+/", FLAGS)
MULTILINE_COMMENT_STRING_START = '/*'


def parse_object(
    s_and_end, strict, scan_once, object_hook, object_pairs_hook,
    memo=None, _w=WHITESPACE.match, _ws=WHITESPACE_STR,
    _ilcs=INLINE_COMMENT_STRING_START, _ilc=INLINE_COMMENT.match,
    _mlcs=MULTILINE_COMMENT_STRING_START, _mlc=MULTILINE_COMMENT.match
):
    '''
    Modified json.decoder.JSONObject function from standard json module
    (python 3.7.7).
    '''
    # pylint: disable=invalid-name, no-else-return, no-else-break
    s, end = s_and_end
    pairs = []
    pairs_append = pairs.append
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = s[end:end + 1]
    # Normally we expect nextchar == '"'
    if nextchar != '"':
        while True:  # Handle comments and whitespaces
            if nextchar in _ws:
                end = _w(s, end).end()
            elif s[end:].startswith(_ilcs):
                end = _ilc(s, end).end()
            elif s[end:].startswith(_mlcs):
                end = _mlc(s, end).end()
            else:
                break
            nextchar = s[end:end + 1]

        # Trivial empty object
        if nextchar == '}':
            if object_pairs_hook is not None:
                result = object_pairs_hook(pairs)
                return result, end + 1
            pairs = {}
            if object_hook is not None:
                pairs = object_hook(pairs)
            return pairs, end + 1
        elif nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end)
    end += 1
    while True:
        key, end = scanstring(s, end, strict)
        key = memo_get(key, key)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end:end + 1] != ':':
            while True:  # Handle comments and whitespaces
                if s[end:end + 1] in _ws:
                    end = _w(s, end).end()
                elif s[end:].startswith(_ilcs):
                    end = _ilc(s, end).end()
                elif s[end:].startswith(_mlcs):
                    end = _mlc(s, end).end()
                else:
                    break
            if s[end:end + 1] != ':':
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            while True:  # Handle comments and whitespaces
                if s[end] in _ws:
                    end = _w(s, end).end()
                elif s[end:].startswith(_ilcs):
                    end = _ilc(s, end).end()
                elif s[end:].startswith(_mlcs):
                    end = _mlc(s, end).end()
                else:
                    break
        except IndexError:
            pass

        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        pairs_append((key, value))

        try:
            nextchar = s[end]
            while True:  # Handle comments and whitespaces
                if nextchar in _ws:
                    end = _w(s, end).end()
                elif s[end:].startswith(_ilcs):
                    end = _ilc(s, end).end()
                elif s[end:].startswith(_mlcs):
                    end = _mlc(s, end).end()
                else:
                    break
                nextchar = s[end]
        except IndexError:
            nextchar = ''
        end += 1

        if nextchar == '}':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)

        while True:  # Handle comments and whitespaces
            if s[end] in _ws:
                end = _w(s, end).end()
            elif s[end:].startswith(_ilcs):
                end = _ilc(s, end).end()
            elif s[end:].startswith(_mlcs):
                end = _mlc(s, end).end()
            else:
                break
        nextchar = s[end:end + 1]
        end += 1
        if nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end - 1)
    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end


def parse_array(
    s_and_end, scan_once, _w=WHITESPACE.match, _ws=WHITESPACE_STR,
    _ilcs=INLINE_COMMENT_STRING_START, _ilc=INLINE_COMMENT.match,
    _mlcs=MULTILINE_COMMENT_STRING_START, _mlc=MULTILINE_COMMENT.match
):
    '''
    Modified json.decoder.JSONArray function from standard module json
    (python 3.7.7).
    '''
    # pylint: disable=invalid-name, no-else-return, no-else-break
    s, end = s_and_end
    values = []
    nextchar = s[end:end + 1]
    while True:  # Handle comments and whitespaces
        if nextchar in _ws:
            end = _w(s, end).end()
        elif s[end:].startswith(_ilcs):
            end = _ilc(s, end).end()
        elif s[end:].startswith(_mlcs):
            end = _mlc(s, end).end()
        else:
            break
        nextchar = s[end:end + 1]

    # Look-ahead for trivial empty array
    if nextchar == ']':
        return values, end + 1
    _append = values.append
    while True:
        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        _append(value)
        nextchar = s[end:end + 1]

        while True:  # Handle comments and whitespaces
            if nextchar in _ws:
                end = _w(s, end).end()
            elif s[end:].startswith(_ilcs):
                end = _ilc(s, end).end()
            elif s[end:].startswith(_mlcs):
                end = _mlc(s, end).end()
            else:
                break
            nextchar = s[end:end + 1]
        end += 1

        if nextchar == ']':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)

        try:
            while True:  # Handle comments and whitespaces
                if s[end] in _ws:
                    end = _w(s, end).end()
                elif s[end:].startswith(_ilcs):
                    end = _ilc(s, end).end()
                elif s[end:].startswith(_mlcs):
                    end = _mlc(s, end).end()
                else:
                    break
        except IndexError:
            pass

    return values, end


class JSONCDecoder(json.JSONDecoder):
    '''
    JSONDecoder with support for C-style comments. Similar to JSONC files from
    Visual Studio code but without support for trailing commas.

    JSONCDecoder can be used in `cls` argument of `json.load` and `json.loads`
    of startdard Python module - `json`.
    '''
    # pylint: disable=arguments-differ

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, *args, **kwargs)
        self.parse_object = parse_object
        self.parse_array = parse_array

        # we need to recreate the internal scan function ..
        self.scan_once = scanner.py_make_scanner(self)

    def decode(
            self, s, _w=WHITESPACE.match,
            _ws=WHITESPACE_STR,
            _ilcs=INLINE_COMMENT_STRING_START, _ilc=INLINE_COMMENT.match,
            _mlcs=MULTILINE_COMMENT_STRING_START,
            _mlc=MULTILINE_COMMENT.match):
        idx = 0
        try:
            while True:  # Handle comments and whitespaces
                if s[idx] in _ws:
                    idx = _w(s, idx).end()
                elif s[idx:].startswith(_ilcs):
                    idx = _ilc(s, idx).end()
                elif s[idx:].startswith(_mlcs):
                    idx = _mlc(s, idx).end()
                else:
                    break
        except IndexError:
            pass
        obj, end = self.raw_decode(s, idx)
        end = _w(s, end).end()
        if end != len(s):
            raise JSONDecodeError("Extra data", s, end)
        return obj




from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

GEOMETRY_BUILD_SCRIPT = '''
-- Geometry
CREATE TABLE GeometryFile (
    GeometryFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX GeometryFile_ResourcePack_fk
ON GeometryFile (ResourcePack_fk);

CREATE TABLE Geometry (
    Geometry_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    GeometryFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    parent TEXT,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (GeometryFile_fk) REFERENCES GeometryFile (GeometryFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX Geometry_GeometryFile_fk
ON Geometry (GeometryFile_fk);
'''

def load_geometries(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for geometry_path in (rp_path / "models").rglob("*.json"):
        load_geometry(db, geometry_path, rp_id)

def load_geometry(db: Connection, geometry_path: Path, rp_id: int):
    cursor = db.cursor()
    # GEOMETRY FILE
    cursor.execute(
        "INSERT INTO GeometryFile (path, ResourcePack_fk) VALUES (?, ?)",
        (geometry_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        geometry_jsonc = load_jsonc(geometry_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    # Try with 1.8.0 format
    for identifier in geometry_jsonc // str:
        full_identifier = cast(str, identifier.parent_key)
        if not full_identifier.startswith("geometry."):
            continue
        # Check for inheritance
        split = full_identifier.split(":", 1)
        id = full_identifier
        parent: Optional[str] = None
        if len(split)  == 2:
            id, parent = split
        cursor.execute(
            '''
            INSERT INTO Geometry (
                identifier, parent, GeometryFile_fk, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (id, parent, file_pk, identifier.path_str)
        )
    # Try with 1.12.0 format
    geometries = (geometry_jsonc / "minecraft:geometry" // int)
    for geometry in geometries:
        identifier = geometry / 'description' / 'identifier'
        identifier_data = identifier.data
        if not isinstance(identifier_data, str):
            continue
        if not identifier_data.startswith("geometry."):
            continue
        cursor.execute(
            '''
            INSERT INTO Geometry (
                identifier, GeometryFile_fk, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (identifier_data, file_pk, geometry.path_str)
        )



from typing import cast
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

RP_ANIMATION_BUILD_SCRIPT = '''
-- RpAnimation
CREATE TABLE RpAnimationFile (
    RpAnimationFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationFile_ResourcePack_fk
ON RpAnimationFile (ResourcePack_fk);

CREATE TABLE RpAnimation (
    RpAnimation_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimationFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    
    FOREIGN KEY (RpAnimationFile_fk) REFERENCES RpAnimationFile (RpAnimationFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimation_RpAnimationFile_fk
ON RpAnimation (RpAnimationFile_fk);

CREATE TABLE RpAnimationParticleEffect (
    RpAnimationParticleEffect_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimation_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RpAnimation_fk) REFERENCES RpAnimation (RpAnimation_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationParticleEffect_RpAnimation_fk
ON RpAnimationParticleEffect (RpAnimation_fk);

CREATE TABLE RpAnimationSoundEffect (
    RpAnimationSoundEffect_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimation_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RpAnimation_fk) REFERENCES RpAnimation (RpAnimation_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationSoundEffect_RpAnimation_fk
ON RpAnimationSoundEffect (RpAnimation_fk);
'''

def load_rp_animations(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for animation_path in (rp_path / "animations").rglob("*.json"):
        load_rp_animation(db, animation_path, rp_id)

def load_rp_animation(db: Connection, animation_path: Path, rp_id: int):
    cursor = db.cursor()
    # RP ANIMATION FILE
    cursor.execute(
        "INSERT INTO RpAnimationFile (path, ResourcePack_fk) VALUES (?, ?)",
        (animation_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        animations_walker = load_jsonc(animation_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    for animation_walker in animations_walker / "animations" // str:
        identifier_data: str = cast(str, animation_walker.parent_key)
        if not identifier_data.startswith("animation."):
            continue
        cursor.execute(
            '''
            INSERT INTO RpAnimation (
                RpAnimationFile_fk, identifier, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (file_pk, identifier_data, animation_walker.path_str)
        )
        rpanim_pk = cursor.lastrowid
        # LOAD PARTICLE EFFECTS
        for particle_effect_walker in (
                animation_walker / "particle_effects" // str):
            short_name = particle_effect_walker / "effect"
            if not isinstance(short_name.data, str):
                continue
            cursor.execute(
                '''
                INSERT INTO RpAnimationParticleEffect (
                    RpAnimation_fk, shortName, jsonPath
                ) VALUES (?, ?, ?)
                ''',
                (rpanim_pk, short_name.data, particle_effect_walker.path_str)
            )
        # LOAD SOUND EFFECTS
        for sound_effect_walker in animation_walker / "sound_effects" // str:
            short_name = sound_effect_walker / "effect"
            if not isinstance(short_name.data, str):
                continue
            cursor.execute(
                '''
                INSERT INTO RpAnimationSoundEffect (
                    RpAnimation_fk, shortName, jsonPath
                ) VALUES (?, ?, ?)
                ''',
                (rpanim_pk, short_name.data, sound_effect_walker.path_str)
            )




from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
from .utils import parse_format_version
import json

BP_ITEM_BUILD_SCRIPT = '''
-- BpItem
CREATE TABLE BpItemFile (
    BpItemFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX BpItemFile_BehaviorPack_fk
ON BpItemFile (BehaviorPack_fk);


CREATE TABLE BpItemParserVersionEnum (
    -- Emulates enum type. Stores possible value of BpItem.parserVersion
    --- column.
    value TEXT PRIMARY KEY
);
INSERT INTO BpItemParserVersionEnum (value) VALUES ('1.10'), ('1.16.100');

CREATE TABLE BpItem (
    BpItem_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BpItemFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    -- The version of the parser. The value is based on fromat_version property
    -- of the file or matching schema (when format_version is missing).
    -- It divides the items into two groups 1.10 and 1.16.100. Most of the
    -- format_versions in Minecraft don't change the format so you need only
    -- two parsers to handle all the items.
    parserVersion TEXT NOT NULL,
    -- The texture of the item. It only exists in 1.16.100+ items.
    texture TEXT,
    
    FOREIGN KEY (BpItemFile_fk) REFERENCES BpItemFile (BpItemFile_pk)
        ON DELETE CASCADE
    -- Constraint emulates enum
    FOREIGN KEY (parserVersion) REFERENCES BpItemParserVersionEnum (value)
);
CREATE INDEX BpItem_BpItemFile_fk
ON BpItem (BpItemFile_fk);
'''

def load_bp_items(db: Connection, bp_id: int):
    bp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (bp_id,)
    ).fetchone()[0]

    for item_path in (bp_path / "items").rglob("*.json"):
        load_bp_item(db, item_path, bp_id)

def load_bp_item(db: Connection, item_path: Path, bp_id: int):
    cursor = db.cursor()
    # BP ITEM FILE
    cursor.execute(
        "INSERT INTO BpItemFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (item_path.as_posix(), bp_id)
    )
    file_pk = cursor.lastrowid
    try:
        item_walker = load_jsonc(item_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    # BP ITEM
    identifier = item_walker / "minecraft:item" / "description" / "identifier"
    if not isinstance(identifier.data, str):
        return  # Silently skip items without an identifier

    format_version_walker = item_walker / "format_version"
    parser_version: Optional[str] = None
    texture: Optional[str] = None
    if isinstance(format_version_walker.data, str):
        try:
            format_version_tuple = parse_format_version(format_version_walker.data)
            if format_version_tuple >= (1, 16, 100):
                parser_version = "1.16.100"
            else:
                parser_version = "1.10"
        except ValueError:
            pass
    # Parse texture if not 1.10
    if parser_version is None or parser_version == "1.16.100":
        texture_walker = (
            item_walker / "minecraft:item" / "components" / "minecraft:icon" /
            "texture")
        if isinstance(texture_walker.data, str):
            texture = texture_walker.data
    # If version unknown, base it on existance of texture
    if parser_version is None:
        if texture is None:
            parser_version = "1.10"
        else:
            parser_version = "1.16.100"
    cursor.execute(
        '''
        INSERT INTO BpItem (
            BpItemFile_fk, identifier, parserVersion, texture
        ) VALUES (?, ?, ?, ?)
        ''',
        (file_pk, identifier.data, parser_version, texture)
    )
    # bp_anim_pk = cursor.lastrowid



import json

class CompactEncoder(json.JSONEncoder):
    '''
    JSONEncoder can be used as `cls` argument to `json.dump` and `json.dumps`.
    It creates formatted JSON strings, with indentation that are more compact
    than the dafault formatting from `json` module. The main difference is that
    the lists of primitives are not split into multiple lines. 
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.indent = -1
        self.respect_indent = True

    def _is_primitive(self, obj):
        return isinstance(obj, (int, bool, str, float))

    def encode(self, obj):
        '''
        Return a JSON string representation of a Python data structure.

        Example:
            >>> CompactEncoder().encode({"foo": ["bar", "baz"]})
            '{\\n\\t"foo": ["bar", "baz"]\\n}'
        '''
        return ''.join([i for i in self.iterencode(obj)])

    def iterencode(self, obj):
        '''
        Encode the given object and yield each string representation line by
        line.

        Example:
            >>> item = {"foo": ["bar", "baz"]}
            >>> ''.join(list(CompactEncoder().iterencode(item))) == \\
            ... CompactEncoder().encode(item)
            True
        '''
        self.indent += 1
        if self.respect_indent:
            ind = self.indent*'\t'
        else:
            ind = ''
        if isinstance(obj, dict):
            if len(obj) == 0:
                yield f"{ind}{{}}"
            else:
                body = []
                for k, v in obj.items():
                    body.extend([
                        f'{j[:self.indent]}"{k}": {j[self.indent:]}'
                        for j in self.iterencode(v)
                    ])
                body_str = ",\n".join(body)
                yield (
                    f'{ind}{{\n'
                    f'{body_str}\n'
                    f'{ind}}}'
                )
        elif isinstance(obj, (list, tuple)):
            primitive_list = True
            for i in obj:
                if not self._is_primitive(i):
                    primitive_list = False
                    break
            if primitive_list:
                body = []
                self.respect_indent = False
                for i in obj:
                    body.extend([j for j in self.iterencode(i)])
                self.respect_indent = True
                yield f'{ind}[{", ".join(body)}]'
            else:
                body = []
                for i in obj:
                    body.extend([j for j in self.iterencode(i)])
                body_str = ",\n".join(body)
                yield (
                    f'{ind}[\n'
                    f'{body_str}\n'
                    f'{ind}]'
                )
        elif self._is_primitive(obj):
            if isinstance(obj, str):
                yield f'{ind}"{str(obj)}"'
            else:
                yield f'{ind}{str(obj).lower()}'
        elif obj is None:
            yield f'{ind}null'
        else:
            raise TypeError('Object of type set is not JSON serializable')
        self.indent -= 1



from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
from .utils import parse_format_version
import json

RP_ITEM_BUILD_SCRIPT = '''
-- RpItem
CREATE TABLE RpItemFile (
    RpItemFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpItemFile_ResourcePack_fk
ON RpItemFile (ResourcePack_fk);


CREATE TABLE RpItem (
    RpItem_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpItemFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    icon TEXT,
    
    FOREIGN KEY (RpItemFile_fk) REFERENCES RpItemFile (RpItemFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpItem_RpItemFile_fk
ON RpItem (RpItemFile_fk);
'''

def load_rp_items(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for item_path in (rp_path / "items").rglob("*.json"):
        load_rp_item(db, item_path, rp_id)

def load_rp_item(db: Connection, item_path: Path, rp_id: int):
    cursor = db.cursor()
    # RP ITEM FILE
    cursor.execute(
        "INSERT INTO RpItemFile (path, ResourcePack_fk) VALUES (?, ?)",
        (item_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        item_walker = load_jsonc(item_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    # RP ITEM
    identifier = item_walker / "minecraft:item" / "description" / "identifier"
    if not isinstance(identifier.data, str):
        return  # Silently skip items without an identifier

    icon: Optional[str] = None
    icon_walker = (
        item_walker / "minecraft:item" / "components" / "minecraft:icon")
    if isinstance(icon_walker.data, str):
        icon = icon_walker.data
    cursor.execute(
        '''
        INSERT INTO RpItem (
            RpItemFile_fk, identifier, icon
        ) VALUES (?, ?, ?)
        ''',
        (file_pk, identifier.data, icon)
    )
    # rp_anim_pk = cursor.lastrowid



from typing import cast
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

BP_ANIMATION_CONTROLLER_BUILD_SCRIPT = '''
-- BpAnimationController
CREATE TABLE BpAnimationControllerFile (
    BpAnimationControllerFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX BpAnimationControllerFile_BehaviorPack_fk
ON BpAnimationControllerFile (BehaviorPack_fk);

CREATE TABLE BpAnimationController (
    BpAnimationController_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BpAnimationControllerFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    
    FOREIGN KEY (BpAnimationControllerFile_fk) REFERENCES BpAnimationControllerFile (BpAnimationControllerFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX BpAnimationController_BpAnimationControllerFile_fk
ON BpAnimationController (BpAnimationControllerFile_fk);

'''

def load_bp_animation_controllers(db: Connection, bp_id: int):
    bp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (bp_id,)
    ).fetchone()[0]

    for ac_path in (bp_path / "animation_controllers").rglob("*.json"):
        load_bp_animation_controller(db, ac_path, bp_id)


def load_bp_animation_controller(db: Connection, animation_controller_path: Path, bp_id: int):
    cursor = db.cursor()
    # BP ANIMATION FILE
    cursor.execute(
        "INSERT INTO BpAnimationControllerFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (animation_controller_path.as_posix(), bp_id)
    )
    file_pk = cursor.lastrowid
    try:
        acs_walker = load_jsonc(animation_controller_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    for ac_walker in acs_walker / "animation_controllers" // str:
        identifier_data: str = cast(str, ac_walker.parent_key)
        if not identifier_data.startswith("controller.animation."):
            continue
        cursor.execute(
            '''
            INSERT INTO BpAnimationController (
                BpAnimationControllerFile_fk, identifier, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (file_pk, identifier_data, ac_walker.path_str)
        )
        # bp_anim_pk = cursor.lastrowid




from typing import cast
from collections import defaultdict
from sqlite3 import Connection
from pathlib import Path
from typing import Literal, NamedTuple
from copy import copy
import json

from .better_json_tools import JSONWalker, load_jsonc
from .utils import find_molang_resources

RENDER_CONTROLLER_BUILD_SCRIPT = '''
-- Render controller
CREATE TABLE RenderControllerFile (
    RenderControllerFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX RenderControllerFile_ResourcePack_fk
ON RenderControllerFile (ResourcePack_fk);

CREATE TABLE RenderController (
    RenderController_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RenderControllerFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RenderControllerFile_fk) REFERENCES RenderControllerFile (RenderControllerFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX RenderController_RenderControllerFile_fk
ON RenderController (RenderControllerFile_fk);

CREATE TABLE RenderControllerTexturesField (
    RenderControllerTexturesField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RenderController_fk INTEGER NOT NULL,

    ownerArray TEXT,
    inOwnerArrayJsonPath TEXT, -- Path to the item in the owner array
    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RenderController_fk) REFERENCES RenderController (RenderController_pk)
        ON DELETE CASCADE
);
CREATE INDEX RenderControllerTexturesField_RenderController_fk
ON RenderControllerTexturesField (RenderController_fk);

CREATE TABLE RenderControllerMaterialsField (
    RenderControllerMaterialsField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RenderController_fk INTEGER NOT NULL,

    ownerArray TEXT,
    inOwnerArrayJsonPath TEXT, -- Path to the item in the owner array
    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    -- The star pattern that matches the bone name
    boneNamePattern TEXT,
    FOREIGN KEY (RenderController_fk) REFERENCES RenderController (RenderController_pk)
        ON DELETE CASCADE
);
CREATE INDEX RenderControllerMaterialsField_RenderController_fk
ON RenderControllerMaterialsField (RenderController_fk);

CREATE TABLE RenderControllerGeometryField (
    RenderControllerGeometryField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RenderController_fk INTEGER NOT NULL,

    ownerArray TEXT,
    inOwnerArrayJsonPath TEXT, -- Path to the item in the owner array
    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RenderController_fk) REFERENCES RenderController (RenderController_pk)
        ON DELETE CASCADE
);
CREATE INDEX RenderControllerGeometryField_RenderController_fk
ON RenderControllerGeometryField (RenderController_fk);
'''


def load_render_controllers(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for geometry_path in (rp_path / "render_controllers").rglob("*.json"):
        load_render_controller(db, geometry_path, rp_id)


class _LoadRcArraysItem(NamedTuple):
    '''
    One item on a list returned by _load_rc_arrays.
    '''
    short_name: str
    json_path: str

def _load_rc_arrays(
        rc: JSONWalker,
        array_type: Literal["geometry", "material", "texture"],
) -> dict[str, list[_LoadRcArraysItem]]:
    '''
    Loads arrays of specified type from a render controller.
    '''
    if array_type == "geometry":
        array_path = rc / "arrays" / "geometries" // str // int
    elif array_type == "material":
        array_path = rc / "arrays" / "materials" // str // int
    elif array_type == "texture":
        array_path = rc / "arrays" / "textures" // str // int
    else:
        raise ValueError(f"Invalid array type {array_type}")
    result = defaultdict(list)
    for obj in array_path:
        if not isinstance(obj.data, str):
            continue
        array_name = cast(str, obj.parent.parent_key)
        array_name = array_name.lower()
        if not array_name.startswith("array."):
            continue
        array_name = array_name[6:]
        values = find_molang_resources(obj.data, [array_type])[array_type]
        path_str = obj.path_str
        result[array_name].extend(
            [_LoadRcArraysItem(v, path_str) for v in values])
    return dict(result)

def load_render_controller(db: Connection, entity_path: Path, rp_id: int):
    cursor = db.cursor()
    # RENDER CONTROLLER FILE
    cursor.execute(
        "INSERT INTO RenderControllerFile (path, ResourcePack_fk) VALUES (?, ?)",
        (entity_path.as_posix(), rp_id))
    file_pk = cursor.lastrowid
    try:
        entity_jsonc = load_jsonc(entity_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    for rc in entity_jsonc / 'render_controllers' // str:
        rc_parent_key = cast(str, rc.parent_key)
        if not rc_parent_key.startswith("controller.render."):
            continue
        cursor.execute(
            '''
            INSERT INTO RenderController (
                identifier, RenderControllerFile_fk, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (rc_parent_key, file_pk, rc.path_str)
        )
        rc_pk = cursor.lastrowid
        # LOAD TEXTURES
        texture_arrays = _load_rc_arrays(rc, "texture")
        for field in rc / "textures" // int:
            if not isinstance(field.data, str):
                continue
            values = find_molang_resources(field.data, ["texture", "array"])
            # Direct access
            for short_name in values["texture"]:
                cursor.execute(
                    '''
                    INSERT INTO RenderControllerTexturesField (
                        RenderController_fk, shortName, jsonPath
                    ) VALUES (?, ?, ?)
                    ''',
                    (rc_pk, short_name, field.path_str)
                )
            # Access through array
            for array_name in values["array"]:
                for texture_reference in texture_arrays.get(array_name, []):
                    cursor.execute(
                        '''
                        INSERT INTO RenderControllerTexturesField (
                            RenderController_fk,
                            shortName,
                            jsonPath,
                            ownerArray,
                            inOwnerArrayJsonPath
                        ) VALUES (?, ?, ?, ?, ?)
                        ''',
                        (
                            rc_pk,
                            texture_reference.short_name,
                            field.path_str,
                            array_name,
                            texture_reference.json_path
                        )
                    )
        # LOAD MATERIALS
        material_arrays = _load_rc_arrays(rc, "material")
        for field in rc / "materials" // int // str:
            if not isinstance(field.data, str):
                continue
            pattern = field.parent_key
            values = find_molang_resources(field.data, ["material", "array"])
            # Direct access
            for short_name in copy(values["material"]):
                cursor.execute(
                    '''
                    INSERT INTO RenderControllerMaterialsField (
                        RenderController_fk, shortName, jsonPath,
                        boneNamePattern
                    ) VALUES (?, ?, ?, ?)
                    ''',
                    (rc_pk, short_name, field.path_str, pattern)
                )
            # Access through array
            for array_name in values["array"]:
                for material_reference in material_arrays.get(array_name, []):
                    cursor.execute(
                        '''
                        INSERT INTO RenderControllerMaterialsField (
                            RenderController_fk,
                            shortName,
                            jsonPath,
                            ownerArray,
                            inOwnerArrayJsonPath,
                            boneNamePattern
                        ) VALUES (?, ?, ?, ?, ?, ?)
                        ''',
                        (
                            rc_pk,
                            material_reference.short_name,
                            field.path_str,
                            array_name,
                            material_reference.json_path,
                            pattern
                        )
                    )
                    
        # LOAD GEOMETRIES
        geo_arrays = _load_rc_arrays(rc, "geometry")
        field = rc / "geometry"
        if isinstance(field.data, str):
            values = find_molang_resources(field.data, ["geometry", "array"])
            # Direct access
            for short_name in copy(values["geometry"]):
                cursor.execute(
                    '''
                    INSERT INTO RenderControllerGeometryField (
                        RenderController_fk, shortName, jsonPath
                    ) VALUES (?, ?, ?)
                    ''',
                    (rc_pk, short_name, field.path_str)
                )
            # Access through array
            for array_name in values["array"]:
                for geometry_reference in geo_arrays.get(array_name, []):
                    cursor.execute(
                        '''
                        INSERT INTO RenderControllerGeometryField (
                            RenderController_fk,
                            shortName,
                            jsonPath,
                            ownerArray,
                            inOwnerArrayJsonPath
                        ) VALUES (?, ?, ?, ?, ?)
                        ''',
                        (
                            rc_pk,
                            geometry_reference.short_name,
                            field.path_str,
                            array_name,
                            geometry_reference.json_path
                        )
                    )



from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
from .utils import split_item_name
import json

TRADE_TABLE_BUILD_SCRIPT = '''
-- Trade Table
CREATE TABLE TradeTableFile (
    TradeTableFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX TradeTableFile_BehaviorPack_fk
ON TradeTableFile (BehaviorPack_fk);

CREATE TABLE TradeTable (
    TradeTable_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    TradeTableFile_fk INTEGER NOT NULL,

    -- Identifier of the trade table (path to the file relative to the behavior
    -- pack root). Unike some other path based identifiers, this one includes
    -- the file extension.
    identifier TEXT NOT NULL,

    FOREIGN KEY (TradeTableFile_fk) REFERENCES TradeTableFile (TradeTableFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX TradeTable_TradeTableFile_fk
ON TradeTable (TradeTableFile_fk);

CREATE TABLE TradeTableItemField (
    -- A reference to an item inside a trade table.

    TradeTableItemField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    TradeTable_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    dataValue INTEGER,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (TradeTable_fk) REFERENCES TradeTable (TradeTable_pk)
        ON DELETE CASCADE
);
CREATE INDEX TradeTableItemField_TradeTable_fk
ON TradeTableItemField (TradeTable_fk);
'''

def load_trade_tables(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for trade_table_path in (rp_path / "trading").rglob("*.json"):
        load_trade_table(db, trade_table_path, rp_path, rp_id)

def load_trade_table(
        db: Connection, trade_table_path: Path, rp_path: Path, rp_id: int):
    cursor = db.cursor()
    # LOOT TABLE FILE
    cursor.execute(
        "INSERT INTO TradeTableFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (trade_table_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        trade_table_jsonc = load_jsonc(trade_table_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    # LOOT TABLE
    identifier = trade_table_path.relative_to(rp_path).as_posix()
    cursor.execute(
        '''
        INSERT INTO TradeTable (
            identifier, TradeTableFile_fk
        ) VALUES (?, ?)
        ''',
        (identifier, file_pk)
    )
    trade_table_pk = cursor.lastrowid

    # LOOT TABLE ITEM FIELDS
    tier_walker = trade_table_jsonc / "tiers" // int
    trade_walker = (
        tier_walker / "groups" // int / "trades" // int +
        tier_walker / "trades" // int
    )
    gives_wants_walker = (
        trade_walker / "gives" // int +
        trade_walker / "wants" // int
    )
    item_walker = (
        gives_wants_walker / "item" +
        gives_wants_walker / "choice" // int / "item"
    )
    for iw in item_walker:
        if not isinstance(iw.data, str):
            continue
        namespace, name, data_value = split_item_name(iw.data)
        cursor.execute(
            '''
            INSERT INTO TradeTableItemField (
                identifier, dataValue, jsonPath, TradeTable_fk
            ) VALUES (?, ?, ?, ?)
            ''',
            (f'{namespace}:{name}', data_value, iw.path_str, trade_table_pk)
        )



from sqlite3 import Connection
from pathlib import Path


TEXTURE_BUILD_SCRIPT = '''
-- Texture
CREATE TABLE TextureFile (
    TextureFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    -- The identifier is the path without extension. This is added to the DB to
    -- make searches easier.
    identifier TEXT NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX TextureFile_ResourcePack_fk
ON TextureFile (ResourcePack_fk);
'''

def load_textures(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for texture_path in (rp_path / "textures").rglob("*.png"):
        load_texture(db, texture_path, rp_path, rp_id)
    for texture_path in (rp_path / "textures").rglob("*.tga"):
        load_texture(db, texture_path, rp_path, rp_id)
    for texture_path in (rp_path / "textures").rglob("*.jpg"):
        load_texture(db, texture_path, rp_path, rp_id)

def load_texture(db: Connection, texture_path: Path, rp_path: Path, rp_id: int):
    cursor = db.cursor()
    # TEXTURE FILE AND ITS IDENTIFIER
    cursor.execute(
        """
        INSERT INTO TextureFile (
            path, identifier, ResourcePack_fk
        ) VALUES (?, ?, ?)
        """,
        (
            texture_path.as_posix(),
            texture_path.relative_to(rp_path).with_suffix("").as_posix(),
            rp_id
        )
    )


from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

ENTITY_BUILD_SCRIPT = '''
-- Behavior pack entity file & content
CREATE TABLE EntityFile (
    EntityFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX EntityFile_BehaviorPack_fk
ON EntityFile (BehaviorPack_fk);

CREATE TABLE Entity (
    Entity_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    EntityFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    FOREIGN KEY (EntityFile_fk) REFERENCES EntityFile (EntityFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX Entity_EntityFile_fk
ON Entity (EntityFile_fk);

CREATE TABLE EntityLootFieldComponentTypeEnum (
    -- This table is used to store possible values for the 
    -- EntityLootField.componentType column. Stores the names of the
    -- components that can have a loot table reference (like 
    -- "minecraft:loot" or "minecraft:equipment")
    value TEXT PRIMARY KEY
);
INSERT INTO EntityLootFieldComponentTypeEnum (value)
VALUES ("minecraft:loot"), ("minecraft:equipment");


CREATE TABLE EntityLootField (
    EntityLootField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Entity_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    componentType TEXT NOT NULL,

    FOREIGN KEY (Entity_fk) REFERENCES Entity (Entity_pk)
        ON DELETE CASCADE
    -- Constraint emulates enum
    FOREIGN KEY (componentType) REFERENCES EntityLootFieldComponentTypeEnum (value)
);
CREATE INDEX EntityLootField_Entity_fk
ON EntityLootField (Entity_fk);

CREATE TABLE EntityTradeFieldComponentTypeEnum (
    -- This table is used to store possible values for the
    -- EntityTradeField.componentType column. Stores the names of the
    -- components that can have a trade table reference (like
    -- "minecraft:economy_trade_table" or "minecraft:trade_table")
    value TEXT PRIMARY KEY
);
INSERT INTO EntityTradeFieldComponentTypeEnum (value)
VALUES ("minecraft:economy_trade_table"), ("minecraft:trade_table");

CREATE TABLE EntityTradeField (
    EntityTradeField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Entity_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    componentType TEXT NOT NULL,

    FOREIGN KEY (Entity_fk) REFERENCES Entity (Entity_pk)
        ON DELETE CASCADE
    -- Constraint emulates enum
    FOREIGN KEY (componentType) REFERENCES EntityTradeFieldComponentTypeEnum (value)
);
CREATE INDEX EntityTradeField_Entity_fk
ON EntityTradeField (Entity_fk);
'''

def load_entities(db: Connection, bp_id: int):
    bp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (bp_id,)
    ).fetchone()[0]

    for entity_path in (bp_path / "entities").rglob("*.json"):
        load_entity(db, entity_path, bp_id)

def load_entity(db: Connection, entity_path: Path, bp_id: int):
    cursor = db.cursor()
    # ENTITY FILE
    cursor.execute(
        "INSERT INTO EntityFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (entity_path.as_posix(), bp_id))

    file_pk = cursor.lastrowid
    try:
        entity_jsonc = load_jsonc(entity_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    entity_walker = entity_jsonc / "minecraft:entity"
    description = entity_walker / "description"

    # ENTITY - IDENTIFIER
    identifier = (description / "identifier").data
    if not isinstance(identifier, str):
        # Skip entitites without identifiers
        return
    cursor.execute(
        '''
        INSERT INTO Entity (
        identifier, EntityFile_fk
        ) VALUES (?, ?)
        ''',
        (identifier, file_pk))
    entity_pk = cursor.lastrowid


    all_componet_groups = (
        entity_walker / "component_groups" // str+
        entity_walker / "components"
    )
    # LOOT
    loot_table_walkers = (
        all_componet_groups / "minecraft:loot" / "table" +
        all_componet_groups / "minecraft:equipment" / "table"
    )
    for loot_table_walker in loot_table_walkers:
        loot_table = loot_table_walker.data
        if not isinstance(loot_table, str):
            continue
        cursor.execute(
            '''
            INSERT INTO EntityLootField (
            identifier, jsonPath, Entity_fk, componentType
            ) VALUES (?, ?, ?, ?)
            ''',
            (
                loot_table,
                loot_table_walker.path_str,
                entity_pk,

                # minecraft:loot OR minecraft:equipment
                loot_table_walker.parent.parent_key
            )
        )

    # TRADE
    trade_table_walkers = (
        all_componet_groups / "minecraft:economy_trade_table" / "table" +
        all_componet_groups / "minecraft:trade_table" / "table"
    )
    for trade_table_walker in trade_table_walkers:
        trade_table = trade_table_walker.data
        if not isinstance(trade_table, str):
            continue
        cursor.execute(
            '''
            INSERT INTO EntityTradeField (
            identifier, jsonPath, Entity_fk, componentType
            ) VALUES (?, ?, ?, ?)
            ''',
            (
                trade_table,
                trade_table_walker.path_str,
                entity_pk,

                # minecraft:trade_table OR minecraft:economy_trade_table
                trade_table_walker.parent.parent_key
            )
        )



from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

LOOT_TABLE_BUILD_SCRIPT = '''
-- Loot Table
CREATE TABLE LootTableFile (
    LootTableFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX LootTableFile_BehaviorPack_fk
ON LootTableFile (BehaviorPack_fk);

CREATE TABLE LootTable (
    LootTable_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    LootTableFile_fk INTEGER NOT NULL,

    -- Identifier of the loot table (path to the file relative to the behavior
    -- pack root). Unike some other path based identifiers, this one includes
    -- the file extension.
    identifier TEXT NOT NULL,

    FOREIGN KEY (LootTableFile_fk) REFERENCES LootTableFile (LootTableFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX LootTable_LootTableFile_fk
ON LootTable (LootTableFile_fk);

CREATE TABLE LootTableItemField (
    -- A reference to an item inside a loot table.

    LootTableItemField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    LootTable_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (LootTable_fk) REFERENCES LootTable (LootTable_pk)
        ON DELETE CASCADE
);
CREATE INDEX LootTableItemField_LootTable_fk
ON LootTableItemField (LootTable_fk);

CREATE TABLE LootTableLootTableField (
    -- A reference to an loot table inside another loot table.

    LootTableLootTableField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    LootTable_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (LootTable_fk) REFERENCES LootTable (LootTable_pk)
        ON DELETE CASCADE
);
CREATE INDEX LootTableLootTableField_LootTable_fk
ON LootTableLootTableField (LootTable_fk);
'''

def load_loot_tables(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for loot_table_path in (rp_path / "loot_tables").rglob("*.json"):
        load_loot_table(db, loot_table_path, rp_path, rp_id)

def load_loot_table(
        db: Connection, loot_table_path: Path, rp_path: Path, rp_id: int):
    cursor = db.cursor()
    # LOOT TABLE FILE
    cursor.execute(
        "INSERT INTO LootTableFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (loot_table_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        loot_table_jsonc = load_jsonc(loot_table_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    # LOOT TABLE
    identifier = loot_table_path.relative_to(rp_path).as_posix()
    cursor.execute(
        '''
        INSERT INTO LootTable (
            identifier, LootTableFile_fk
        ) VALUES (?, ?)
        ''',
        (identifier, file_pk)
    )
    loot_table_pk = cursor.lastrowid

    # LOOT TABLE ITEM FIELDS
    entry_walker = loot_table_jsonc / "pools" // int / "entries" // int
    while len(entry_walker) > 0:
        for ew in entry_walker:
            ew_name = ew / "name"
            if not isinstance(ew_name.data, str):
                continue
            ew_type = ew / "type"
            if ew_type.data == "item":
                cursor.execute(
                    '''
                    INSERT INTO LootTableItemField (
                        identifier, jsonPath, LootTable_fk
                    ) VALUES (?, ?, ?)
                    ''',
                    (ew_name.data, ew_name.path_str, loot_table_pk)
                )
            elif ew_type.data == "loot_table":
                cursor.execute(
                    '''
                    INSERT INTO LootTableLootTableField (
                        identifier, jsonPath, LootTable_fk
                    ) VALUES (?, ?, ?)
                    ''',
                    (ew_name.data, ew_name.path_str, loot_table_pk)
                )
        # Entry property can have pools. This is a nested structure.
        entry_walker = entry_walker / "pools" // int / "entries" // int



from sqlite3 import Connection
from pathlib import Path
from typing import Union

BEHAVIOR_PACK_BUILD_SCRIPT = '''
-- Behavior Pack
CREATE TABLE BehaviorPack (
    BehaviorPack_pk INTEGER PRIMARY KEY AUTOINCREMENT,

    path Path NOT NULL
);
CREATE INDEX BehaviorPack_path
ON BehaviorPack (path);
'''

def load_behavior_pack(db: Connection, rp_path: Union[Path, str]) -> int:
    if isinstance(rp_path, Path):
        rp_path = rp_path.as_posix()
    count = db.execute(
        "SELECT total(1) FROM BehaviorPack WHERE path = ?",
        (rp_path,)).fetchone()[0]
    if count != 0:
        raise ValueError("RP already loaded.")
    cursor = db.cursor()
    cursor.execute(
        "INSERT INTO BehaviorPack (path) VALUES (?)",
        (rp_path,))
    return cursor.lastrowid  # type: ignore



from typing import cast
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

BP_ANIMATION_BUILD_SCRIPT = '''
-- BpAnimation
CREATE TABLE BpAnimationFile (
    BpAnimationFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BehaviorPack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (BehaviorPack_fk) REFERENCES BehaviorPack (BehaviorPack_pk)
        ON DELETE CASCADE
);
CREATE INDEX BpAnimationFile_BehaviorPack_fk
ON BpAnimationFile (BehaviorPack_fk);

CREATE TABLE BpAnimation (
    BpAnimation_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    BpAnimationFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    
    FOREIGN KEY (BpAnimationFile_fk) REFERENCES BpAnimationFile (BpAnimationFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX BpAnimation_BpAnimationFile_fk
ON BpAnimation (BpAnimationFile_fk);

'''

def load_bp_animations(db: Connection, bp_id: int):
    bp_path: Path = db.execute(
        "SELECT path FROM BehaviorPack WHERE BehaviorPack_pk = ?",
        (bp_id,)
    ).fetchone()[0]

    for animation_path in (bp_path / "animations").rglob("*.json"):
        load_bp_animation(db, animation_path, bp_id)

def load_bp_animation(db: Connection, animation_path: Path, bp_id: int):
    cursor = db.cursor()
    # BP ANIMATION FILE
    cursor.execute(
        "INSERT INTO BpAnimationFile (path, BehaviorPack_fk) VALUES (?, ?)",
        (animation_path.as_posix(), bp_id)
    )
    file_pk = cursor.lastrowid
    try:
        animations_walker = load_jsonc(animation_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    for animation_walker in animations_walker / "animations" // str:
        identifier_data: str = cast(str, animation_walker.parent_key)
        if not identifier_data.startswith("animation."):
            continue
        cursor.execute(
            '''
            INSERT INTO BpAnimation (
                BpAnimationFile_fk, identifier, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (file_pk, identifier_data, animation_walker.path_str)
        )
        # bp_anim_pk = cursor.lastrowid



from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

PARTICLE_BUILD_SCRIPT = '''
-- Particle
CREATE TABLE ParticleFile (
    ParticleFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX ParticleFile_ResourcePack_fk
ON ParticleFile (ResourcePack_fk);

CREATE TABLE Particle (
    Particle_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ParticleFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    material TEXT,
    texture TEXT,
    FOREIGN KEY (ParticleFile_fk) REFERENCES ParticleFile (ParticleFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX Particle_ParticleFile_fk
ON Particle (ParticleFile_fk);
'''

def load_particles(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for particle_path in (rp_path / "particles").rglob("*.json"):
        load_particle(db, particle_path, rp_id)

def load_particle(db: Connection, particle_path: Path, rp_id: int):
    cursor = db.cursor()
    # PARTICLE FILE
    cursor.execute(
        "INSERT INTO ParticleFile (path, ResourcePack_fk) VALUES (?, ?)",
        (particle_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        particle_walker = load_jsonc(particle_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    description = particle_walker / "particle_effect" / "description"
    basic_render_parameters = description / "basic_render_parameters"

    identifier = description / "identifier"
    material = basic_render_parameters / "material"
    texture = basic_render_parameters / "texture"

    identifier_data = identifier.data
    if not isinstance(identifier_data, str):
        return
    material_data = material.data
    if not isinstance(material.data, str):
        material_data = None
    texture_data = texture.data
    if not isinstance(texture.data, str):
        texture_data = None
    cursor.execute(
        '''
        INSERT INTO Particle (
            identifier, material, texture, ParticleFile_fk
        ) VALUES (?, ?, ?, ?)
        ''',
        (identifier_data, material_data, texture_data, file_pk)
    )
